/*---------------------------------------------------------------------------

 Copyright (c) 2005
	by Dirk Meyer, All rights reserved.
	Im Grund 4, 34317 Habichtswald, Germany
	Email: dirk.meyer@dinoex.sub.org

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. Neither the name of the author nor the names of any co-contributors
    may be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

-----------------------------------------------------------------------------

	shell - a plugin for xchat2
	===========================

	$Id: shell.c,v 1.5 2006/07/25 09:34:05 dm Exp $

----------------------------------------------------------------------------*/

#include "xchat-plugin.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>

#include <fcntl.h>
#ifdef linux
#include <sys/file.h>
#include <db_185.h>
#else
#include <db.h>
#endif


int xchat_plugin_init(xchat_plugin *plugin_handle,
	const char **plugin_name,
	const char **plugin_desc,
	const char **plugin_version,
	const char *arg);
void xchat_plugin_deinit(void);


#define	MAX_BUF	1024

typedef struct _queue_entry {
	struct _queue_entry *q_next;
	char *q_data;
} queue_entry;

typedef struct _shell_entry {
	char *s_trigger;
	char *s_command;
} shell_entry;


static xchat_plugin *ph; /* plugin handle */
static xchat_hook *timerhook = NULL;
static int enable = 1;
static int shell_count = 0;
static char *configdir = NULL;
static char *shell_config = NULL;
static char *watched_channel = NULL;
static queue_entry *notice_queue = NULL;
static shell_entry shell_list[ MAX_BUF ];
static char fbuf[MAX_BUF];
static char qbuf[MAX_BUF];
static char nbuf[MAX_BUF];
static char cbuf[MAX_BUF];


static queue_entry *
queue_init(const char *data)
{
	queue_entry *new;

	new = malloc(sizeof(queue_entry));
	if (new == NULL) {
		xchat_printf(ph, "new queue, out of mem\n");
		return NULL;
	}
	new->q_next = NULL;
	new->q_data = (data != NULL) ? strdup(data) : NULL;
	return new;
}

static void
queue_add(queue_entry **queue, const char *data)
{
	queue_entry *old;
	queue_entry *new;

	if (queue == NULL)
		return;

	new = queue_init(data);
	if (*queue == NULL) {
		*queue = new;
		return;
	}
	old = *queue;
	while (old->q_next != NULL)
		old = old->q_next;
	old->q_next = new;
}

static char *
queue_get(queue_entry **queue)
{
	queue_entry *old;
	char *data;

	if (queue == NULL)
		return NULL;

	old = *queue;
	if (old == NULL)
		return NULL;

	*queue = old->q_next;
	data = old->q_data;
	free(old);
	return data;
}

static void
queue_delete(queue_entry **queue)
{
	queue_entry *old;

	if (queue == NULL)
		return;

	old = *queue;
	if (old == NULL)
		return;

	while (old != NULL) {
		if (old->q_data != NULL)
			free(old->q_data);
		*queue = old->q_next;
		free(old);
		old = *queue;
	}
	return;
}

static char *
queue_addf(queue_entry **queue, const char *format, ...)
#ifdef __GNUC__
__attribute__ ((format(printf,2,3)))
#endif
;

static char *
queue_addf(queue_entry **queue, const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	vsnprintf(qbuf, sizeof(qbuf), format, ap);
	va_end(ap);
	queue_add(queue, qbuf);
	return qbuf;
}

static void
clean_arg( char *arg)
{
	char *cptr;
	char ch;

	cptr = arg;
	while ((ch = (*cptr)) != 0) {
		if  (( ch >= '0' ) && ( ch <= '9' )) {
			cptr++;
			continue;
		}
		if  (( ch >= 'a' ) && ( ch <= 'z' )) {
			cptr++;
			continue;
		}
		if  (( ch >= 'A' ) && ( ch <= 'Z' )) {
			cptr++;
			continue;
		}
		if  (( ch == ' ' ) || ( ch == '/' )) {
			cptr++;
			continue;
		}
		*cptr = '.';
		cptr++;
	}
}

static void
shell_clean(void)
{
	int i;

	if (watched_channel != NULL) {
		free(watched_channel);
		watched_channel = NULL;
	}
	for ( i = 0; i <= shell_count; i++ ) {
		free(shell_list[ i ].s_trigger);
		free(shell_list[ i ].s_command);
	}
	shell_count = 0;
}

static int
shell_update(void)
{
	FILE *fp;
	char *data;
	char *nl;
	char *cmd;
	int found = 0;

	shell_clean();

	fp = fopen(shell_config, "r");
	if (fp == NULL) {
		xchat_printf(ph, "shell cant read: %s\n", shell_config);
		return found;
	}
	while (!feof(fp)) {
		data = fgets(fbuf, sizeof(fbuf) -1, fp);
		if (data == NULL) {
			/* no data */
			break;
		}
		nl = strchr(data, '\n');
		if (nl != NULL) *nl = 0;
		if ((found++) == 0) {
			watched_channel = strdup( data );
			continue;
		}
		if (found >= MAX_BUF) {
			/* to much data */
			 break;
		}
		cmd = strchr(data, '|');
		if (cmd == NULL)
			continue;
		*(cmd++) = 0;
		xchat_printf(ph, "shell trigger <%s> action: %s\n", data, cmd);
		shell_list[ shell_count ].s_trigger = strdup( data );
		shell_list[ shell_count ].s_command = strdup( cmd );
		shell_count ++;
	}
	fclose(fp);
	return found;
}

static int
timeout_cb(void *userdata)
{
	xchat_context *backup;
	xchat_context *watch;
	char *data;

	/* process queued buffer */
	if (notice_queue != NULL) {
		backup = xchat_get_context(ph);
		watch = xchat_find_context(ph, NULL, watched_channel);
		if (watch != NULL) {
			data = queue_get(&notice_queue);
			if (data != NULL) {
				xchat_set_context(ph, watch);
				xchat_command(ph, data);
				free(data);
			};
		};
		xchat_set_context(ph, backup);
	};
	return 1; /* return 1 to keep the timeout going */
}

#if 0
	word	text

	1	:dinoex!~dinoex@euirc-c679c75c.dinoex.sub.de
	2	PRIVMSG
	3	dinoex
	4	:xdcc
	5	list | send
	6	group | pack
#endif

static void
extract_nick(const char *from)
{
	size_t len;
	char *end;

	nbuf[0] = 0;
	if (from == NULL)
		return;
	if (from[0] == 0)
		return;
	from ++;
	len = strlen(from);
	if ((len <= 0) || (len >= sizeof(nbuf))) {
		xchat_printf(ph, "oversized nick %s\n", from);
		return;
	}
	strncpy(nbuf, from, len);
	nbuf[sizeof(nbuf) - 1] = 0;

	end = strchr(nbuf, '!' );
	if ( end == NULL )
		return;

	*end = 0;
	return;
}

static int
shell_open(const char *command, const char *arg)
{
	FILE *fp;
	char *arg_clean;
	char *data;
	char *nl;

	if (arg == NULL)
		arg = "";
	arg_clean = strdup(arg);
	clean_arg(arg_clean);
	snprintf(cbuf, sizeof(cbuf), "%s '%s' < /dev/null", command, arg_clean);
	free(arg_clean);

	fp = popen(cbuf, "r");
	if ( fp == NULL ) {
		xchat_printf(ph, "failed: %s\n", cbuf);
		return 1;
	}
	while (!feof(fp)) {
		data = fgets(fbuf, sizeof(fbuf) -1, fp);
		if (data == NULL) {
			/* no data */
			break;
		}
		nl = strchr(data, '\n');
		if (nl != NULL) *nl = 0;
		queue_addf(&notice_queue, "notice %s %s", nbuf, data);
        }
	return pclose(fp);
}

static void
shell_parse(char *word[], char *word_eol[], void *userdata)
{
	int i;

	for ( i = 0; i < shell_count; i++ ) {
		if (strcasecmp(word[4], shell_list[ i ].s_trigger) == 0) {
			extract_nick(word[1]);
			shell_open(shell_list[ i ].s_command, word_eol[5]);
			return;
		}
	}
}

static int
privmsg_cb(char *word[], char *word_eol[], void *userdata)
{
	int i;
	const char *str;

	if (enable == 0)
		return XCHAT_EAT_NONE; /* don't eat this event, let other plugins and xchat see it too */

	if (word[4] == NULL)
		return XCHAT_EAT_NONE; /* don't eat this event, let other plugins and xchat see it too */

	str = NULL;
	if (xchat_get_prefs (ph, "irc_nick1", &str, &i) == 1) {
		if (xchat_nickcmp(ph,word[3], str) != 0) {
			shell_parse(word, word_eol, userdata);
			return XCHAT_EAT_NONE; /* don't eat this event, let other plugins and xchat see it too */
		}
	}
	if (watched_channel != NULL ) {
		if (strcasecmp(word[3], watched_channel) != 0) {
			shell_parse(word, word_eol, userdata);
			return XCHAT_EAT_NONE; /* don't eat this event, let other plugins and xchat see it too */
		}
	}
        return XCHAT_EAT_NONE; /* don't eat this event, let other plugins and xchat see it too */
}

static int
shellupdate_cb(char *word[], char *word_eol[], void *userdata)
{
	shell_update();
	return XCHAT_EAT_ALL; /* eat this command so xchat and other plugins can't process it */
}

static void
timer_end(void)
{
	if (timerhook == NULL)
		return;

	xchat_unhook(ph, timerhook);
	timerhook = NULL;
}

static void
timer_start(void)
{
	if (timerhook == NULL)
		timerhook = xchat_hook_timer(ph, 1000, timeout_cb, NULL);
}

static int
shell_toggle_cb(char *word[], char *word_eol[], void *userdata)
{
	if (enable == 0) {
		enable = 1;
		timer_start();
		xchat_print(ph, "shell now enabled!\n");
		shell_update();
	} else {
		enable = 0;
		queue_delete(&notice_queue);
		timer_end();
		xchat_print(ph, "shell now disabled!\n");
	}

	return XCHAT_EAT_ALL; /* eat this command so xchat and other plugins can't process it */
}

int
xchat_plugin_init(xchat_plugin *plugin_handle,
	const char **plugin_name,
	const char **plugin_desc,
	const char **plugin_version,
	const char *arg)
{
	const char *cptr;

	/* we need to save this for use with any xchat_* functions */
	ph = plugin_handle;

	*plugin_name = "shell";
	*plugin_desc = "shell extensions";
	*plugin_version = "0.2";

	cptr = xchat_get_info(ph, "xchatdirfs");
	if (cptr == NULL)
		cptr = ".";
	configdir = strdup(cptr);
	snprintf(fbuf, sizeof(fbuf), "%s/shell.config", configdir);
	shell_config = strdup(fbuf);

	xchat_hook_command(ph, "shell", XCHAT_PRI_NORM, shell_toggle_cb,
		"Usage: shell, Turns OFF/ON shell", 0);
	xchat_hook_command(ph, "shellupdate", XCHAT_PRI_NORM, shellupdate_cb,
		"Usage: shellupdate, read config", 0);

	xchat_hook_server(ph, "PRIVMSG", XCHAT_PRI_NORM, privmsg_cb, NULL);

	timerhook = NULL;
	timer_start();

	xchat_printf(ph, "%s Plugin %s loaded successfully!\n", *plugin_name, *plugin_version);
	shell_update();

	return 1; /* return 1 for success */
}

void
xchat_plugin_deinit(void)
{
	queue_delete(&notice_queue);
	if (configdir != NULL)
		free(configdir);
	if (shell_config != NULL)
		free(shell_config);
	shell_clean();
}

/* EOF */
